import hashlib
import os
import traceback
from datetime import datetime, timedelta

import sys
from slack_sdk import WebClient as SlackWebClient


class SlackApi:
    def __init__(self, project_name: str, slack_token: str, slack_channel_id: str):
        self.client = SlackWebClient(slack_token)
        self.project_name = project_name
        self.slack_channel_id = slack_channel_id

    def post_message(self, message: str, error_text: str = None):
        """ Post ordinary messages as warning or error messages as danger """
        attachments = {
            "text": message,
            "fallback": message,
            "color": "warning"
        }

        if error_text:
            pre_text = f"[{self.project_name}]: {message}"
            attachments.update({
                "text": error_text,
                "pretext": pre_text,
                "fallback": pre_text,
                "title": "Error traceback",
                "color": "danger"
            })

        self.client.chat_postMessage(channel=self.slack_channel_id,
                                     attachments=[attachments],
                                     username=f"{self.project_name.lower()}-logger",
                                     icon_emoji=":robot_face:")


class SlackLogger:
    def __init__(self, project_name: str, slack_token: str, slack_channel_id: str) -> None:
        self.project_name = project_name
        self.slacker = SlackApi(
            project_name=project_name,
            slack_token=slack_token,
            slack_channel_id=slack_channel_id
        )
        self.last_messages = []

    @staticmethod
    def hash_error(error_text):
        return hashlib.md5(error_text.encode()).hexdigest()

    def save_and_post_to_slack(self, error_text, message):
        self.last_messages.append({
            'date': datetime.now().replace(microsecond=0),
            'message_hash': self.hash_error(error_text)
        })

        truncated_error_text = error_text[-8000:] if len(error_text) > 8000 else error_text
        self.slacker.post_message(message=message, error_text=truncated_error_text)
        # create logs directory if not exists
        if not os.path.exists("logs"):
            os.makedirs("logs")
        with open(f"logs/{self.project_name}.log", "w", encoding='utf-8') as file:
            file.write(f'\n \n DATE: {datetime.now().replace(microsecond=0)}: ERROR in {message} \n \n')
            e_type, e_val, e_tb = sys.exc_info()
            traceback.print_exception(e_type, e_val, e_tb, file=file)

    def error(self, exc: Exception, header_message: str, error_additional_data=None) -> None:
        """
        :param exc: Exception object appears as red text in slack.
        :param header_message: bold text appears above the error message - usually the place where the error occurred
        :param error_additional_data: Additional data to be added to the error message like variables, etc.
        :return: None
        """
        error_text = ''.join(traceback.format_exception(None, exc, exc.__traceback__))
        if error_additional_data:
            error_text += f"\n\n\nAdditional data:\n{error_additional_data}"
        error_text_hash = self.hash_error(error_text)

        two_minutes_ago = datetime.now() - timedelta(minutes=2)
        self.last_messages = [item for item in self.last_messages if item.get('date') > two_minutes_ago]

        last_messages_hashes = [item.get('message_hash') for item in self.last_messages]

        if error_text_hash in last_messages_hashes:
            return

        self.save_and_post_to_slack(error_text, header_message)

    def info(self, message: str) -> None:
        """
        :param message: message appears as a warning message in slack without error
        :return: None
        """
        self.slacker.post_message(message=message)



import asyncio
import logging
import warnings
from functools import wraps
from typing import Callable, get_type_hints

from utils_b_infra.caching import BaseCache, CacheConfig, key_builder

__all__ = ["Cache", "CacheConfig"]

logger: logging.Logger = logging.getLogger(__name__)
logger.addHandler(logging.NullHandler())


def get_typed_return_annotation(func) -> type | None:
    """
    Returns the type annotations of the given function.

    Parameters:
    func (function): The function whose type annotations are to be retrieved.

    Returns:
    type: The type annotations of the given function.
    """
    return_type = get_type_hints(func)
    return return_type.get('return')


class Cache:
    def __init__(self, config: CacheConfig = CacheConfig()) -> None:
        self._config = config
        self._coder = self._config.coder
        self._backend_cache = self._get_cache_backend()

    def _get_cache_backend(self) -> BaseCache:
        cache_type = self._config.cache_type
        if cache_type == 'SimpleCache':
            warnings.warn("SimpleCache is not recommended for production environment with multiple workers.")
            from utils_b_infra.caching.backends import SimpleCache
            return SimpleCache(
                threshold=self._config.simple_cache_threshold,
                default_timeout=self._config.default_timeout
            )

        if cache_type == 'RedisCache':
            from utils_b_infra.caching.backends import RedisCache
            return RedisCache(
                url=self._config.redis_url,
                host=self._config.redis_host,
                port=self._config.redis_port,
                password=self._config.redis_password,
                db=self._config.redis_db,
                app_space=self._config.app_space,
                default_timeout=self._config.default_timeout
            )

        if cache_type == 'MongoCache':
            from utils_b_infra.caching.backends import MongoCache
            return MongoCache(
                url=self._config.mongo_url,
                database=self._config.mongo_database,
                collection=self._config.mongo_collection,
                direct_connection=self._config.mongo_direct_connection,
                app_space=self._config.app_space,
                default_timeout=self._config.default_timeout
            )

    def _get_is_sliding_expiration(self, endpoint_sliding_expiration: bool | None) -> bool:
        """
        Check if the endpoint has sliding expiration enabled, if yes, return its value over the global config.
        """
        if endpoint_sliding_expiration is None and self._config.sliding_expiration:
            return True
        return endpoint_sliding_expiration or False

    @staticmethod
    async def ensure_async_func(func, *args, **kwargs):
        """
        Ensures that the given function runs asynchronously.
        """
        if asyncio.iscoroutinefunction(func):
            return await func(*args, **kwargs)
        else:
            loop = asyncio.get_event_loop()
            from functools import partial
            func_part = partial(func, *args, **kwargs)
            return await loop.run_in_executor(None, func_part)

    def cached(self,
               timeout: int = None,
               sliding_expiration: bool = None,
               namespace: str = "") -> Callable[[Callable], Callable]:
        """
        Decorator to cache the result of a function.
        :param timeout : Timeout in seconds (int).
                  Set to `0` to never expire. If not specified, the default timeout from the cache config is used.
                  A pre-calculated values in the cache_config can be used, e.g., `cache_config.ONE_HOUR`,
                  `cache_config.ONE_DAY`, etc.
        :param sliding_expiration : Enable sliding expiration for the cache key (bool).
                  sliding window expiration mechanism resets the expiration time on every access
                  to the cache key, this means that the cache key will only expire if it is not accessed
                  for the duration of the timeout.

                  * sliding_expiration at the endpoint level takes precedence over the global config.

        :param namespace: Namespace for the cache keys.
        """

        def decorator(func: Callable):
            @wraps(func)
            async def async_wrapper(*args, **kwargs):
                # Key building and sliding expiration logic stays the same
                is_sliding_expiration = self._get_is_sliding_expiration(sliding_expiration)
                key = await key_builder(
                    func,
                    app_space=self._config.app_space,
                    namespace=namespace,
                    args=args,
                    kwargs=kwargs
                )

                # Try to retrieve the cache
                try:
                    ttl, cached_result = await self.get_with_ttl(key)
                except Exception as e:
                    logger.warning(
                        f"Error retrieving cache key '{key}' from backend: {e}",
                        exc_info=True,
                    )
                    ttl, cached_result = 0, None

                if cached_result is not None:  # Cache hit
                    result = self._coder.decode(cached_result)
                    if is_sliding_expiration:
                        await self.set(key, cached_result, timeout)
                else:  # Cache miss
                    result = await self.ensure_async_func(func, *args, **kwargs)
                    result_encoded = self._coder.encode(result)
                    await self.set(key, result_encoded, timeout)

                return result

            @wraps(func)
            def sync_wrapper(*args, **kwargs):
                """Convert synchronous function to asynchronous by running in the event loop"""
                try:
                    loop = asyncio.get_event_loop()
                    if loop.is_running():
                        # If the loop is already running, create a new task for async_wrapper
                        return asyncio.ensure_future(async_wrapper(*args, **kwargs))
                    else:
                        # If the loop is not running, run async_wrapper in the current loop
                        return loop.run_until_complete(async_wrapper(*args, **kwargs))
                except RuntimeError:
                    # If no event loop is available, create a new one and run async_wrapper
                    return asyncio.run(async_wrapper(*args, **kwargs))

            if asyncio.iscoroutinefunction(func):
                return async_wrapper
            else:
                return sync_wrapper

        return decorator

    async def get_with_ttl(self, key: str) -> tuple[int, bytes | None]:
        return await self._backend_cache.get_with_ttl(key)

    async def get(self, key: str) -> bytes | None:
        return await self._backend_cache.get(key)

    async def set(self, key: str, value: bytes, expire: int = None) -> None:
        await self._backend_cache.set(key, value, expire)

    async def clear(self, namespace: str = None, key: str = None) -> int:
        return await self._backend_cache.clear(namespace, key)



from typing import Literal

from utils_b_infra.generic import retry_with_timeout


class TextTranslator:
    """
    To use google translate, set GOOGLE_APPLICATION_CREDENTIALS env variable
    to the path of the google service account json file before using this class:
    os.environ['GOOGLE_APPLICATION_CREDENTIALS'] = join(current_path, 'google_service_account.json')

    Requires:
        the utils-b-infra package with the translation extras:
        pip install utils-b-infra[translation]

    Parameters:
        deepl_api_key (str): API key for DeepL.
        languages (dict[str, str]): Supported language mappings.
        google_project_id (str): Google Cloud project ID.
    """

    def __init__(self, languages: dict[str, str], deepl_api_key: str = None, google_project_id: str = None):
        if not deepl_api_key and not google_project_id:
            raise ValueError("Either 'deepl_api_key' or 'google_project_id' must be provided.")

        self.languages = languages
        self.deepl_translate_client = None
        self.google_translate_client = None

        if deepl_api_key:
            import deepl
            self.deepl_translate_client = deepl.Translator(deepl_api_key)

        if google_project_id:
            try:
                from google.cloud.translate_v3 import TranslationServiceClient as GoogleTranslationServiceClient
                from google.cloud.translate_v3.types import TranslateTextRequest
            except ImportError:
                raise ImportError(
                    "Please install the translation dependencies with: pip install utils-b-infra[translation]")
            self.TranslateTextRequest = TranslateTextRequest
            self.google_translate_client = GoogleTranslationServiceClient()
            self.google_project_id = google_project_id

    @retry_with_timeout(retries=3, timeout=120, initial_delay=20, backoff=2)
    def _translate_text_with_google(self,
                                    text_to_translate: str,
                                    source_language: str,
                                    target_lang: str,
                                    mime_type: str) -> str:
        """Translates text into the target language.

        Target must be an ISO 639-1 language code.
        See https://cloud.google.com/translate/docs/languages for the supported languages
        """

        # Prepare the request
        request = self.TranslateTextRequest(
            parent=f"projects/{self.google_project_id}",
            contents=[text_to_translate],
            mime_type=mime_type,  # mime types: text/plain, text/html
            source_language_code=source_language,
            target_language_code=target_lang,
        )

        # Perform the translation
        response = self.google_translate_client.translate_text(request=request)

        # Extract the translated text
        translated_text = response.translations[0].translated_text

        return translated_text

    @retry_with_timeout(retries=2, timeout=120, initial_delay=30, backoff=2)
    def _translate_text_with_deepl(self,
                                   text_to_translate: str,
                                   source_lang: str,
                                   target_lang: str,
                                   tag_handling: str = None):
        result = self.deepl_translate_client.translate_text(
            text=text_to_translate,
            source_lang=source_lang.upper(),
            target_lang=target_lang,
            tag_handling=tag_handling
        )
        # The result is a list of TextResult objects, or a single TextResult object if the text was one string.
        if isinstance(result, list):
            return result[0].text
        return result.text

    def get_translations(self,
                         text_: str,
                         replace_lang_url: bool = False,
                         source_language: Literal["en", "ru", "ar", "de", "es", "fr", "uk"] = "en",
                         target_langs: list[Literal["ru", "ar", "de", "es", "fr", "uk"]] = None,
                         engine: Literal["google", "deepl"] = "google",
                         google_mime_type: Literal["text/plain", "text/html"] = "text/plain",
                         deepl_tag_handling: str = None) -> dict[str, str]:
        """
        Translate text to all languages in LANGUAGES dict
        :param text_: text to translate
        :param replace_lang_url: replace base url with the base url from target language in the text.
        :param source_language: source language ['en', 'ru', 'ar', 'de', 'es', 'fr', 'uk']
        :param target_langs: list of languages to translate to, if None, translate to all languages in LANGUAGES dict
        :param engine: the translation engine to use, either "google" or "deepl"
        :param google_mime_type: for Google only, mime type of the text, either "text/plain" or "text/html"
        :param deepl_tag_handling: for deepl only, how to handle tags in the text, either "xml" or "html"
        :return: dict with translations
        """
        if engine and engine not in ["google", "deepl"]:
            raise ValueError("engine must be either 'google' or 'deepl'")
        if source_language not in ['en', 'ru', 'ar', 'de', 'es', 'fr', 'uk']:
            raise ValueError("source_language must be one of 'en', 'ru', 'ar', 'de', 'es', 'fr', 'uk'")
        if target_langs and not all(lang in self.languages for lang in target_langs):
            raise ValueError("target_langs must be a list of 'ru', 'ar', 'de', 'es', 'fr', 'uk'")

        translations = {'en': text_}
        if not text_:
            return translations

        for lang in self.languages:
            if target_langs and lang not in target_langs:
                continue
            if engine == "google":
                translated_text = self._translate_text_with_google(
                    text_to_translate=text_,
                    source_language=source_language,
                    target_lang=lang,
                    mime_type=google_mime_type
                )
            elif engine == "deepl":
                translated_text = self._translate_text_with_deepl(
                    text_to_translate=text_,
                    source_lang=source_language,
                    target_lang=lang,
                    tag_handling=deepl_tag_handling
                )
            else:
                raise ValueError("engine must be either 'google' or 'deepl'")

            if replace_lang_url:
                translated_text = translated_text.replace('https://us-uk.bookimed.com/',
                                                          self.languages[lang])

            if lang == 'uk':
                lang = 'ua'

            translations[lang] = translated_text

        return translations
